use crate::{
    echo_to_term_frame,
    id_intern::intern_id,
    match_table::{create_match_table, match_table_add, match_table_remove},
    GlobalLayoutCell, GlobalLayoutState, GlobalMemoCell, TermFrame,
};
use gc_arena::{Gc, Rootable};
use itertools::Itertools;
use piccolo::{
    self, async_sequence, Callback, CallbackReturn, Context, Error, FromValue, Function, IntoValue,
    SequenceReturn, StashedTable, StashedUserData, StashedValue, Table, UserData, Value, Variadic,
};
use std::{rc::Rc, str};
use yew::UseStateSetter;

use super::call_checking_metatable;

pub fn alias(ctx: Context<'_>) -> Callback<'_> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let info: Table = ctx.get_global("info")?;
        let cur_frame_id: TermFrame =
            try_unwrap_frame(ctx, &info.get(ctx, ctx.intern_static(b"current_frame"))?)?;
        let frames: Table = ctx.get_global("frames")?;
        let cur_frame: Table = frames.get(ctx, cur_frame_id.0 as i64)?;
        let aliases: UserData = cur_frame.get(ctx, "aliases")?;

        if stack.is_empty() {
            return list_match_tab(cur_frame_id, aliases, ctx);
        }

        let alias_match: piccolo::String = piccolo::String::from_value(
            ctx,
            stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("Missing alias match"))?,
        )?;
        let sub_to: piccolo::String = piccolo::String::from_value(
            ctx,
            stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("Missing substitution match"))?,
        )?;
        if !stack.is_empty() {
            Err(anyhow::Error::msg(
                "Extra arguments to alias command. Try wrapping the action in {}",
            ))?;
        }

        stack.push_back(aliases.into_value(ctx));
        stack.push_back(alias_match.into_value(ctx));
        stack.push_back(sub_to.into_value(ctx));
        let seq = async_sequence(&ctx, |locals, mut seq| {
            let add_func = locals.stash(&ctx, Function::Callback(match_table_add(ctx)));
            async move {
                seq.call(&add_func, 0).await?;
                Ok(SequenceReturn::Return)
            }
        });

        Ok(piccolo::CallbackReturn::Sequence(seq))
    })
}

pub fn unalias(ctx: Context<'_>) -> Callback<'_> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let info: Table = ctx.get_global("info")?;
        let cur_frame_id: TermFrame =
            try_unwrap_frame(ctx, &info.get(ctx, ctx.intern_static(b"current_frame"))?)?;
        let frames: Table = ctx.get_global("frames")?;
        let cur_frame: Table = frames.get(ctx, cur_frame_id.0 as i64)?;
        let aliases: UserData = cur_frame.get(ctx, "aliases")?;

        let alias_match: piccolo::String = piccolo::String::from_value(
            ctx,
            stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("Missing alias match"))?,
        )?;
        if !stack.is_empty() {
            Err(anyhow::Error::msg(
                "Extra arguments to unalias command. Repeating the substitution is not necessary.",
            ))?;
        }

        stack.push_back(aliases.into_value(ctx));
        stack.push_back(alias_match.into_value(ctx));
        let seq = async_sequence(&ctx, |locals, mut seq| {
            let remove_func = locals.stash(&ctx, Function::Callback(match_table_remove(ctx)));
            async move {
                seq.call(&remove_func, 0).await?;
                Ok(SequenceReturn::Return)
            }
        });

        Ok(piccolo::CallbackReturn::Sequence(seq))
    })
}

pub fn list_match_tab<'gc>(
    frame: TermFrame,
    aliases: UserData<'gc>,
    ctx: Context<'gc>,
) -> Result<CallbackReturn<'gc>, Error<'gc>> {
    let seq = async_sequence(&ctx, |locals, mut seq| {
        let frame = locals.stash(&ctx, intern_id(ctx, frame));
        let aliases = locals.stash(&ctx, aliases);
        async move {
            call_checking_metatable::<StashedUserData>(&mut seq, aliases, "lua_table", &[]).await?;
            let echo_func = seq.try_enter(|ctx, locals, _ex, mut stack| {
                let alias_tab: Table = stack.consume(ctx)?;
                stack.push_back(locals.fetch(&frame).into_value(ctx));
                stack.push_back(
                    alias_tab
                        .iter()
                        .map(|(k, v)| {
                            Ok::<String, Error>(format!(
                                "{} => {}",
                                &String::from_value(ctx, k)?,
                                &String::from_value(ctx, v)?
                            ))
                        })
                        .collect::<Result<Vec<String>, Error>>()?
                        .iter()
                        .join("\r\n")
                        .into_value(ctx),
                );
                Ok(locals.stash(
                    &ctx,
                    ctx.get_global::<Table>("commands")?
                        .get::<&str, Function>(ctx, "echo_frame")?,
                ))
            })?;

            seq.call(&echo_func, 0).await?;

            Ok(SequenceReturn::Return)
        }
    });
    Ok(CallbackReturn::Sequence(seq))
}

pub fn echo_frame_raw<'gc, 'a>(
    ctx: Context<'gc>,
    global_memo: &'a GlobalMemoCell,
    _global_layout: &'a UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let frame: TermFrame = try_unwrap_frame(
            ctx,
            &stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("Missing frame"))?,
        )?;
        let message: piccolo::String = stack.from_front(ctx)?;
        let message_str = str::from_utf8(message.as_bytes())
            .map_err(|_| "Expected message to echo to be UTF-8.".into_value(ctx))?;
        echo_to_term_frame(&global_memo, &frame, message_str).map_err(|m| m.into_value(ctx))?;
        Ok(piccolo::CallbackReturn::Return)
    })
}

pub fn echo_frame<'gc>(
    ctx: Context<'gc>,
    _global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let commands: Table<'gc> = ctx.get_global("commands")?;
        let function: Function = commands.get(ctx, "echo_frame_raw")?;
        let frame_no: Value = stack
            .pop_front()
            .ok_or_else(|| anyhow::Error::msg("Missing frame number"))?;
        let all_parts: Vec<String> = stack
            .consume::<Variadic<Vec<Value>>>(ctx)?
            .into_iter()
            .map(|v| format!("{}", v.display()))
            .collect();
        stack.push_front(frame_no);
        let message = ctx.intern((all_parts.join(" ") + "\r\n").as_bytes());
        stack.push_back(message.into());
        Ok(piccolo::CallbackReturn::Call {
            function,
            then: None,
        })
    })
}

pub fn echo<'gc>(
    ctx: Context<'gc>,
    _global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let commands: Table<'gc> = ctx.get_global("commands")?;
        let function: Function = commands.get(ctx, "echo_frame")?;
        let info: Table<'gc> = ctx.get_global("info")?;
        let cur_frame: Value = info.get(ctx, ctx.intern_static(b"current_frame"))?;
        stack.push_front(cur_frame);
        Ok(piccolo::CallbackReturn::Call {
            function,
            then: None,
        })
    })
}

pub fn vsplit<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let frame: TermFrame = TermFrame(stack.from_front(ctx)?);
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .vsplit(&path, frame.clone())
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(CallbackReturn::Call {
            function: Function::Callback(ensure_frame_instance(ctx, &frame)),
            then: None,
        })
    })
}

pub fn hsplit<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let frame: TermFrame = TermFrame(stack.from_front(ctx)?);
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .hsplit(&path, frame.clone())
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(CallbackReturn::Call {
            function: Function::Callback(ensure_frame_instance(ctx, &frame)),
            then: None,
        })
    })
}

pub fn panel_merge<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .join(&path)
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(piccolo::CallbackReturn::Return)
    })
}

pub fn try_unwrap_frame<'gc>(
    ctx: Context<'gc>,
    value: &Value<'gc>,
) -> Result<TermFrame, piccolo::Error<'gc>> {
    match u64::from_value(ctx, *value) {
        Ok(v) => Ok(TermFrame(v)),
        Err(_) => Ok(UserData::from_value(ctx, *value)?
            .downcast::<Rootable!['gcb => Gc<'gcb, TermFrame>]>()?
            .as_ref()
            .clone()),
    }
}

pub fn ensure_frame_instance<'gc>(ctx: Context<'gc>, frame: &TermFrame) -> Callback<'gc> {
    let frame = frame.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, _stack| {
        let frames: Table = ctx.get_global("frames")?;
        let frame_value: Value = frames.get(ctx, frame.0 as i64)?;
        if !frame_value.is_nil() {
            return Ok(CallbackReturn::Return);
        }

        let frame_tab = Table::new(&ctx);
        let classes: Table = ctx.get_global("classes")?;
        frame_tab.set_metatable(&ctx, Some(classes.get(ctx, "frame")?));
        frames.set(ctx, frame.0 as i64, frame_tab)?;

        // Call frame_tab:new(frame) to setup.
        let frame = intern_id::<TermFrame>(ctx, frame.clone());
        let seq = async_sequence(&ctx, move |locals, mut seq| {
            let frame_tab = locals.stash(&ctx, frame_tab);
            let frame = locals.stash(&ctx, frame);
            async move {
                call_checking_metatable::<StashedTable>(&mut seq, frame_tab, "new", &[frame])
                    .await?;
                Ok(SequenceReturn::Return)
            }
        });
        Ok(CallbackReturn::Sequence(seq))
    })
}

pub(super) fn new_frame<'gc>(ctx: Context<'gc>, _global_memo: &GlobalMemoCell) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let frame_tab: Table = Table::from_value(
            ctx,
            stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("classes.frame:new missing object!"))?,
        )?;
        let frame: Value = stack
            .pop_front()
            .ok_or_else(|| anyhow::Error::msg("classes.frame:new missing frame!"))?;

        frame_tab.set(ctx, ctx.intern_static(b"frame"), frame)?;

        let seq = async_sequence(&ctx, |locals, mut seq| {
            let frame_tab = locals.stash(&ctx, frame_tab);
            let create_match_tab = locals.stash(&ctx, Function::Callback(create_match_table(ctx)));
            async move {
                seq.call(&create_match_tab, 0).await?;
                seq.try_enter(|ctx, locals, _exec, mut stack| {
                    let frame_tab = locals.fetch(&frame_tab);
                    frame_tab.set(ctx, "aliases", stack.consume::<Value>(ctx)?)?;
                    Ok(())
                })?;
                Ok(piccolo::SequenceReturn::Return)
            }
        });

        Ok(piccolo::CallbackReturn::Sequence(seq))
    })
}

// Note: send_command_to_frame is called from the command parser, and can't be overriden (or even
// called) by the user.
// It will normally result in a call into frame_input, but that can be overriden by scripts.
pub(super) fn send_command_to_frame<'gc>(ctx: Context<'gc>) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let (frame, line): (Table, piccolo::String) = stack.consume(ctx)?;
        let seq = async_sequence(&ctx, |locals, mut seq| {
            let frame = locals.stash(&ctx, frame);
            let line = locals.stash(&ctx, line.into_value(ctx));
            async move {
                call_checking_metatable::<StashedTable>(&mut seq, frame, "input", &[line]).await?;
                Ok(SequenceReturn::Return)
            }
        });
        Ok(CallbackReturn::Sequence(seq))
    })
}

pub(super) fn frame_input<'gc>(ctx: Context<'gc>, _global_memo: &GlobalMemoCell) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let callback_return = {
            let frame_tab: Table = Table::from_value(
                ctx,
                stack
                    .pop_front()
                    .ok_or_else(|| anyhow::Error::msg("classes.frame:input missing object!"))?,
            )?;
            let line: Value = stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("classes.frame:input missing line!"))?;
            stack.consume(ctx)?;

            let aliases: UserData = frame_tab.get(ctx, "aliases")?;
            let frame: Value = frame_tab.get(ctx, "frame")?;

            let seq = async_sequence(&ctx, |locals, mut seq| {
                let aliases = locals.stash(&ctx, aliases);
                let line = locals.stash(&ctx, line);
                let frame_tab = locals.stash(&ctx, frame_tab);
                let frame = locals.stash(&ctx, frame);
                async move {
                    call_checking_metatable::<StashedUserData>(
                        &mut seq,
                        aliases.clone(),
                        "try_run_sub",
                        &[line.clone(), frame],
                    )
                    .await?;
                    let (alias_hit, linked_mud, linked_mud_nil) =
                        seq.try_enter(|ctx, locals, _ex, mut stack| {
                            let linked_mud = locals
                                .fetch(&frame_tab)
                                .get_value(ctx, ctx.intern_static(b"linked_mud"));
                            Ok((
                                stack.consume::<bool>(ctx)?,
                                locals.stash(&ctx, linked_mud),
                                linked_mud.is_nil(),
                            ))
                        })?;

                    if alias_hit || linked_mud_nil {
                        return Ok(SequenceReturn::Return);
                    }

                    call_checking_metatable::<StashedValue>(
                        &mut seq,
                        linked_mud,
                        "mudinput_line",
                        &[line],
                    )
                    .await?;

                    Ok(SequenceReturn::Return)
                }
            });

            Ok(piccolo::CallbackReturn::Sequence(seq))
        };
        callback_return
    })
}
