use wasm_bindgen::JsValue;
use web_sys::console;

#[derive(Eq, PartialEq, Debug)]
pub enum TelnetOutput {
    Line(Vec<u8>),
    Prompt(Vec<u8>), // Like a line but without a newline.
    Nop,
    Break,
    Sync,
    Interrupt,
    AbortOutput,
    AreYouThere,
    Will(u8),
    Wont(u8),
    Do(u8),
    Dont(u8),
    Subnegotiation(Vec<u8>),
}

const IAC: u8 = 255;
const NOP: u8 = 241;
const DATA_MARK: u8 = 242;
const BREAK: u8 = 243;
const INTERRUPT: u8 = 244;
const ABORT: u8 = 245;
const AYT: u8 = 246;
const ERASECHAR: u8 = 247;
const ERASELINE: u8 = 248;
const GOAHEAD: u8 = 249;
const EOR: u8 = 239;
const STARTSUB: u8 = 250;
const ENDSUB: u8 = 240;
const WILL: u8 = 251;
const WONT: u8 = 252;
const DO: u8 = 253;
const DONT: u8 = 254;

pub fn parse_telnet_buf(input: &[u8]) -> (Vec<u8>, Option<TelnetOutput>) {
    let mut ptr: &[u8] = input;
    let mut textbuf: Vec<u8> = vec![];
    loop {
        match ptr.first() {
            None => return (textbuf.to_owned(), None),
            Some(b'\n') => {
                textbuf.push(b'\n');
                return (ptr[1..].to_owned(), Some(TelnetOutput::Line(textbuf)));
            }
            Some(&IAC) => {
                ptr = &ptr[1..];
                match ptr.first() {
                    None => {
                        textbuf.push(IAC);
                        return (textbuf, None);
                    }
                    Some(&NOP) => {}
                    Some(&DATA_MARK) => {}
                    Some(&BREAK) => {
                        textbuf.extend_from_slice(&ptr[1..]);
                        return (textbuf, Some(TelnetOutput::Break));
                    }
                    Some(&INTERRUPT) => {
                        textbuf.extend_from_slice(&ptr[1..]);
                        return (textbuf, Some(TelnetOutput::Interrupt));
                    }
                    Some(&ABORT) => {
                        textbuf.extend_from_slice(&ptr[1..]);
                        return (textbuf, Some(TelnetOutput::AbortOutput));
                    }
                    Some(&AYT) => {
                        textbuf.extend_from_slice(&ptr[1..]);
                        return (textbuf, Some(TelnetOutput::AreYouThere));
                    }
                    Some(&ERASECHAR) => {
                        if !textbuf.is_empty() {
                            textbuf = textbuf[0..textbuf.len() - 1].to_owned();
                        }
                    }
                    Some(&ERASELINE) => {
                        textbuf = vec![];
                    }
                    Some(&GOAHEAD) => {
                        if textbuf.len() > 1 {
                            return (ptr[1..].to_owned(), Some(TelnetOutput::Prompt(textbuf)));
                        }
                    }
                    Some(&EOR) => {
                        if textbuf.len() > 1 {
                            return (ptr[1..].to_owned(), Some(TelnetOutput::Prompt(textbuf)));
                        }
                    }
                    Some(&WILL) => {
                        ptr = &ptr[1..];
                        match ptr.first() {
                            None => {
                                textbuf.push(IAC);
                                textbuf.push(WILL);
                                return (textbuf, None);
                            }
                            Some(c) => {
                                textbuf.extend_from_slice(&ptr[1..]);
                                return (textbuf, Some(TelnetOutput::Will(*c)));
                            }
                        }
                    }
                    Some(&WONT) => {
                        ptr = &ptr[1..];
                        match ptr.first() {
                            None => {
                                textbuf.push(IAC);
                                textbuf.push(WONT);
                                return (textbuf, None);
                            }
                            Some(c) => {
                                textbuf.extend_from_slice(&ptr[1..]);
                                return (textbuf, Some(TelnetOutput::Wont(*c)));
                            }
                        }
                    }
                    Some(&DO) => {
                        ptr = &ptr[1..];
                        match ptr.first() {
                            None => {
                                textbuf.push(IAC);
                                textbuf.push(DO);
                                return (textbuf, None);
                            }
                            Some(c) => {
                                textbuf.extend_from_slice(&ptr[1..]);
                                return (textbuf, Some(TelnetOutput::Do(*c)));
                            }
                        }
                    }
                    Some(&DONT) => {
                        ptr = &ptr[1..];
                        match ptr.first() {
                            None => {
                                textbuf.push(IAC);
                                textbuf.push(DONT);
                                return (textbuf, None);
                            }
                            Some(c) => {
                                textbuf.extend_from_slice(&ptr[1..]);
                                return (textbuf, Some(TelnetOutput::Dont(*c)));
                            }
                        }
                    }
                    Some(&STARTSUB) => {
                        match (1..ptr.len() - 1).find(|i| ptr[*i] == IAC && ptr[*i + 1] == ENDSUB) {
                            None => {
                                // Including the STARTSUB...
                                textbuf.push(IAC);
                                textbuf.extend_from_slice(ptr);
                                return (textbuf, None);
                            }
                            Some(end_idx) => {
                                textbuf.extend_from_slice(&ptr[end_idx + 2..]);
                                return (
                                    textbuf,
                                    Some(TelnetOutput::Subnegotiation(ptr[1..end_idx].to_owned())),
                                );
                            }
                        }
                    }
                    Some(c) => {
                        // Completely unexpected command. Warn and ignore.
                        console::log_1(&JsValue::from_str(&format!(
                            "Received unknown IAC command {}, assuming single byte and ignoring.",
                            c
                        )));
                    }
                }
            }
            Some(c) => textbuf.push(*c),
        }
        ptr = &ptr[1..];
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn parse_telnet_works() {
        assert_eq!(parse_telnet_buf(b"Hello"), (b"Hello".to_vec(), None));
        assert_eq!(
            parse_telnet_buf(b"Hello\r\n"),
            (
                b"".to_vec(),
                Some(TelnetOutput::Line(b"Hello\r\n".to_vec()))
            )
        );
        assert_eq!(
            parse_telnet_buf(b"Hello\r\nWorld"),
            (
                b"World".to_vec(),
                Some(TelnetOutput::Line(b"Hello\r\n".to_vec()))
            )
        );
        assert_eq!(
            parse_telnet_buf(b"Hello> \xff\xf9World"), // Go Ahead after prompt.
            (
                b"World".to_vec(),
                Some(TelnetOutput::Prompt(b"Hello> ".to_vec()))
            )
        );
        assert_eq!(
            parse_telnet_buf(b"Hello> \xff\xefWorld"), // End Of Record after prompt.
            (
                b"World".to_vec(),
                Some(TelnetOutput::Prompt(b"Hello> ".to_vec()))
            )
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xf1World\r\nWoo!"), // NOP in the middle.
            (
                b"Woo!".to_vec(),
                Some(TelnetOutput::Line(b"Hello World\r\n".to_vec()))
            )
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xf2World\r\nWoo!"), // DATA MARK in the middle.
            (
                b"Woo!".to_vec(),
                Some(TelnetOutput::Line(b"Hello World\r\n".to_vec()))
            )
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xf3World\r\nWoo!"), // BREAK in the middle.
            (b"Hello World\r\nWoo!".to_vec(), Some(TelnetOutput::Break))
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xf4World\r\nWoo!"), // INTERRUPT in the middle.
            (
                b"Hello World\r\nWoo!".to_vec(),
                Some(TelnetOutput::Interrupt)
            )
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff"), // Incomplete IAC command.
            (b"Hello \xff".to_vec(), None)
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xf5World!"), // Abort
            (b"Hello World!".to_vec(), Some(TelnetOutput::AbortOutput))
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xf6World!"), // Are You There
            (b"Hello World!".to_vec(), Some(TelnetOutput::AreYouThere))
        );
        assert_eq!(
            parse_telnet_buf(b"Hello B\xff\xf7World!"), // Erase Char
            (b"Hello World!".to_vec(), None)
        );
        assert_eq!(
            parse_telnet_buf(b"Whoops \xff\xf8Hello World!"), // Erase Line
            (b"Hello World!".to_vec(), None)
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xfb\x01World!"), // Will
            (b"Hello World!".to_vec(), Some(TelnetOutput::Will(1)))
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xfc\x02World!"), // Wont
            (b"Hello World!".to_vec(), Some(TelnetOutput::Wont(2)))
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xfd\x03World!"), // Do
            (b"Hello World!".to_vec(), Some(TelnetOutput::Do(3)))
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xfe\x04World!"), // Dont
            (b"Hello World!".to_vec(), Some(TelnetOutput::Dont(4)))
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xfablah"), // Partial subnegotiation
            (b"Hello \xff\xfablah".to_vec(), None)
        );
        assert_eq!(
            parse_telnet_buf(b"Hello \xff\xfablah\xff\xf0World!"), // Partial subnegotiation
            (
                b"Hello World!".to_vec(),
                Some(TelnetOutput::Subnegotiation(b"blah".to_vec()))
            )
        );
    }
}
